/*-----------------------------------------------------------------------*
  File........: protea.p
  Description : A super procedure containing procedures and functions that
                implement the Tiny Encryption Algorithm
  Input Param : <none>
  Output Param: <none>
  Author......: TEA was invented by by David Wheeler and Roger Needham of
                the Cambridge Computer Laboratory.  It was implemented in Progress
                4GL by Brian Tawney (btawney@cornerstar.com).
  Created.....: 09/24/2003
  Notes.......: If you make changes to this program that others could benefit
                from, please share your code with me: btawney@cornerstar.com.
                
                TEA is a Feistel cipher with XOR and addition as the non-linear
                mixing functions.

                PUBLIC FUNCTIONS:
                
                encipher(string, key) returns character
                    This function takes a key of up to 16 characters and a
                    string of any length and returns an enciphered
                    representation of the string.

                decipher(string, key) returns character
                    This function takes a key of up to 16 characters and
                    a representation of an enciphered string and
                    returns the plaintext string.
 
                set_mix(integer) returns logical
                    This function sets the mix variable, if possible.  It
                    returns true if a valid mix value was passed in,
                    otherwise it returns false.  The mix variable is the
                    number of iterations over which to encrypt the data.
                    This number may be from 1 to 32, and is set to 32 by default.
                    The bigger the number, the slower the encryption process
                    but the better the encryption.
 *-----------------------------------------------------------------------*/

/*--CONSTANTS-------------------------------------------------------------*/

/*--FORWARD DEFINITIONS---------------------------------------------------*/

/*  Encipher a string and return an encoded hexadecimal string */
FUNCTION encipher RETURNS CHARACTER (CHARACTER, CHARACTER) FORWARD.

/*  Decipher an encoded hexadecimal string and return the decoded string. */
FUNCTION decipher RETURNS CHARACTER (CHARACTER, CHARACTER) FORWARD.

/*  Set the mix value. */
FUNCTION set_mix RETURNS LOGICAL (INTEGER) FORWARD.

/*  Exclusive-or two integers. */
FUNCTION xor RETURNS INTEGER PRIVATE (INTEGER, INTEGER) FORWARD.

/*  Convert raw data to string-safe data. */
FUNCTION safe RETURNS CHARACTER PRIVATE (RAW) FORWARD.

/*  Convert string-safe data to raw. */
FUNCTION unsafe RETURNS RAW PRIVATE (CHARACTER) FORWARD.

/*--VARIABLE DEFINITIONS--------------------------------------------------*/

/*  This is the initial value of the sum variable in the decipher procedure.
    Its value is calculated in the initialize internal procedure */
DEFINE VARIABLE sum-initial AS INTEGER NO-UNDO.

/*  This is the value of the delta variable (0x9E3779B9).  It is the real part
    of Sqrt(5/4) - 1/2 ~ 0.618034 multiplied by 2^32. */
DEFINE VARIABLE delta AS INTEGER INITIAL -1640531527 NO-UNDO.

/*  mix is the number of iterations to encrypt the data with.  32 is ample,
    16 is sufficient, as few as eight may be OK. The algorithm achieves good
    dispersion after six iterations.  The more iterations, the slower it
    runs.
    
    You MUST use the same mix to decode a string as was used to encode it */
DEFINE VARIABLE mix AS INTEGER INITIAL 32 NO-UNDO.

/*  This is a safe set of 64 characters for converting raw encrypted data
    into a string that is safe for storing in a Progress character field */
DEFINE VARIABLE base64 AS CHARACTER CASE-SENSITIVE NO-UNDO INITIAL
    "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ+-".

/*--PERSISTENT PROCEDURE INITIALIZATION-----------------------------------*/

/*  If this procedure is not running persistent, then raise an error and
    exit */
IF NOT THIS-PROCEDURE:PERSISTENT
THEN RETURN ERROR "Super procedure protea.p was called non-persistently".

/*  Try to make this a super-procedure.  If another instance is already running
    as a super-procedure, then exit. */
RUN make_self_super IN THIS-PROCEDURE NO-ERROR.

IF ERROR-STATUS:ERROR
THEN RETURN.

RUN initialize.

/*--PROCEDURE AND FUNCTION DEFINITIONS-----------------------------------*/

PROCEDURE make_self_super:
    DEFINE VARIABLE i AS INTEGER NO-UNDO.
    DEFINE VARIABLE h AS HANDLE NO-UNDO.
    
    DO i = 1 TO NUM-ENTRIES(SESSION:SUPER-PROCEDURES):
        ASSIGN h = WIDGET-HANDLE(ENTRY(i,SESSION:SUPER-PROCEDURES)).
        
        IF h:FILE-NAME = THIS-PROCEDURE:FILE-NAME
        THEN RETURN ERROR.
    END.
    
    SESSION:ADD-SUPER-PROCEDURE(THIS-PROCEDURE).
END PROCEDURE.

PROCEDURE initialize PRIVATE:
    DEFINE VARIABLE i AS INTEGER NO-UNDO.

    /*  Calculate the initial value of the sum variable for the decipher
        function */
    ASSIGN sum-initial = 0.
    DO i = 1 TO mix:
        ASSIGN sum-initial = sum-initial + delta.
    END.
END PROCEDURE.

/*  Set the mix value. */
FUNCTION set_mix RETURNS LOGICAL (in-mix AS INTEGER):
    IF in-mix >= 1
    AND in-mix <= 64
    THEN DO:
        ASSIGN mix = in-mix.
        
        RUN initialize.
        
        RETURN TRUE.
    END.
    ELSE RETURN FALSE.
END FUNCTION.

/*  Encipher a string and return an encoded hexadecimal string */
FUNCTION encipher RETURNS CHARACTER
    (
    in-string AS CHARACTER,
    in-key AS CHARACTER
    ):
    /*  First, transfer the string and the key into raw variables, so we can
        manipulate the strings easily as four-byte integers. Then calculate
        the result, storing it in a string output variable.
    */
    DEFINE VARIABLE text-length AS INTEGER NO-UNDO.
    
    DEFINE VARIABLE raw-input AS RAW NO-UNDO.
    DEFINE VARIABLE raw-output AS RAW NO-UNDO.
    DEFINE VARIABLE raw-key AS RAW NO-UNDO.
    DEFINE VARIABLE string-output AS CHARACTER NO-UNDO.
    DEFINE VARIABLE dword AS INTEGER NO-UNDO.
    
    DEFINE VARIABLE cy AS INTEGER NO-UNDO.
    DEFINE VARIABLE cz AS INTEGER NO-UNDO.
    DEFINE VARIABLE csum AS INTEGER NO-UNDO.
    DEFINE VARIABLE ca AS INTEGER NO-UNDO.
    DEFINE VARIABLE cb AS INTEGER NO-UNDO.
    DEFINE VARIABLE cc AS INTEGER NO-UNDO.
    DEFINE VARIABLE cd AS INTEGER NO-UNDO.
    DEFINE VARIABLE cn AS INTEGER NO-UNDO.
    
    ASSIGN text-length = 8 * (TRUNCATE(LENGTH(in-string, "RAW") / 8,0) + 1).

    ASSIGN
        LENGTH(raw-input) = text-length
        LENGTH(raw-output) = text-length
        LENGTH(raw-key) = 16
        .

    PUT-STRING(raw-input,1) = in-string.
    PUT-STRING(raw-key,1) = in-key.
    
    ASSIGN
        ca = GET-LONG(raw-key,1)
        cb = GET-LONG(raw-key,5)
        cc = GET-LONG(raw-key,9)
        cd = GET-LONG(raw-key,13)
        .

    DO dword = 1 TO text-length BY 8:
        ASSIGN
            cy = GET-LONG(raw-input,dword)
            cz = GET-LONG(raw-input,dword + 4)
            csum = 0.

        DO cn = 1 TO mix:
            /*  Here's the C code:
                  sum += delta;
                  y += (z<<4)+a ^ z+sum ^ (z>>5)+b;
                  z += (y<<4)+c ^ y+sum ^ (y>>5)+d;
            */
            ASSIGN
                csum = csum + delta
                cy = cy + xor(xor((cz * 16 + ca),(cz + csum)),(GET-BITS(cz,6,27) + cb))
                cz = cz + xor(xor((cy * 16 + cc),(cy + csum)),(GET-BITS(cy,6,27) + cd))
                .
        END.
        
        PUT-LONG(raw-output,dword) = cy.
        PUT-LONG(raw-output,dword + 4) = cz.
    END.

    RETURN safe(raw-output).
END FUNCTION.

/*  Encipher a string and return an encoded hexadecimal string */
FUNCTION decipher RETURNS CHARACTER
    (
    in-string AS CHARACTER,
    in-key AS CHARACTER
    ):
    /*  First, transfer the string and the key into raw variables, so we can
        manipulate the strings easily as four-byte integers. Then calculate
        the result, storing it in a raw output variable.  Then convert the
        raw output to something that can safely be stored in a Progress
        character field without losing any information.
    */
    DEFINE VARIABLE text-length AS INTEGER NO-UNDO.
    
    DEFINE VARIABLE raw-input AS RAW NO-UNDO.
    DEFINE VARIABLE raw-output AS RAW NO-UNDO.
    DEFINE VARIABLE raw-key AS RAW NO-UNDO.
    DEFINE VARIABLE string-output AS CHARACTER NO-UNDO.
    DEFINE VARIABLE dword AS INTEGER NO-UNDO.
    
    DEFINE VARIABLE cy AS INTEGER NO-UNDO.
    DEFINE VARIABLE cz AS INTEGER NO-UNDO.
    DEFINE VARIABLE csum AS INTEGER NO-UNDO.
    DEFINE VARIABLE ca AS INTEGER NO-UNDO.
    DEFINE VARIABLE cb AS INTEGER NO-UNDO.
    DEFINE VARIABLE cc AS INTEGER NO-UNDO.
    DEFINE VARIABLE cd AS INTEGER NO-UNDO.
    DEFINE VARIABLE cn AS INTEGER NO-UNDO.
    
    ASSIGN raw-input = unsafe(in-string).

    ASSIGN
        text-length = LENGTH(raw-input)
        LENGTH(raw-output) = text-length
        LENGTH(raw-key) = 16
        .

    PUT-STRING(raw-key,1) = in-key.
    
    ASSIGN
        ca = GET-LONG(raw-key,1)
        cb = GET-LONG(raw-key,5)
        cc = GET-LONG(raw-key,9)
        cd = GET-LONG(raw-key,13)
        .

    DO dword = 1 TO text-length BY 8:
        ASSIGN
            cy = GET-LONG(raw-input,dword)
            cz = GET-LONG(raw-input,dword + 4)
            csum = sum-initial.
    
        DO cn = 1 TO mix:
            /*  Here's the C code:
                  z -= (y<<4)+c ^ y+sum ^ (y>>5)+d;
                  y -= (z<<4)+a ^ z+sum ^ (z>>5)+b;
                  sum -= delta;
            */
            ASSIGN
                cz = cz - xor(xor((cy * 16 + cc),(cy + csum)),(GET-BITS(cy,6,27) + cd))
                cy = cy - xor(xor((cz * 16 + ca),(cz + csum)),(GET-BITS(cz,6,27) + cb))
                csum = csum - delta
                .
        END.
        
        PUT-LONG(raw-output,dword) = cy.
        PUT-LONG(raw-output,dword + 4) = cz.
    END.

    RETURN GET-STRING(raw-output,1).
END FUNCTION.

/*  Exclusive-or two integers. */
FUNCTION xor RETURNS INTEGER PRIVATE (a AS INTEGER, b AS INTEGER):
    DEFINE VARIABLE c AS INTEGER NO-UNDO.
    DEFINE VARIABLE i AS INTEGER NO-UNDO.
    
    DO i = 1 TO 32:
        PUT-BITS(c,i,1) = GET-BITS(a,i,1) * GET-BITS(b,i,1).
    END.
    
    RETURN c.
END FUNCTION.

/*  Convert integer data to string-safe data */
FUNCTION safe RETURNS CHARACTER PRIVATE (in-raw AS RAW):
    DEFINE VARIABLE out-string AS CHARACTER CASE-SENSITIVE NO-UNDO.
    DEFINE VARIABLE b AS INTEGER NO-UNDO.
    DEFINE VARIABLE sub AS INTEGER NO-UNDO.
    
    /*  Creating a safe string will mean inflating the string.
        We will save as much space as possible by converting to base 64.
    */

    DO b = 1 TO LENGTH(in-raw) BY 4:
        ASSIGN
            sub = GET-LONG(in-raw,b)
            out-string = out-string
                + SUBSTRING(base64,GET-BITS(sub,31,2) + 1,1)
                + SUBSTRING(base64,GET-BITS(sub,25,6) + 1,1)
                + SUBSTRING(base64,GET-BITS(sub,19,6) + 1,1)
                + SUBSTRING(base64,GET-BITS(sub,13,6) + 1,1)
                + SUBSTRING(base64,GET-BITS(sub,7,6) + 1,1)
                + SUBSTRING(base64,GET-BITS(sub,1,6) + 1,1)
            .
    END.

    RETURN out-string.
END FUNCTION.

/*  Convert string-safe data to integer */
FUNCTION unsafe RETURNS RAW PRIVATE (in-string AS CHARACTER):
    DEFINE VARIABLE out-raw AS RAW NO-UNDO.
    DEFINE VARIABLE b AS INTEGER NO-UNDO.
    DEFINE VARIABLE sub AS INTEGER NO-UNDO.
    DEFINE VARIABLE insert-at AS INTEGER NO-UNDO.
    
    /*  This reverses the safe function */
    ASSIGN LENGTH(out-raw) = INTEGER(TRUNCATE(LENGTH(in-string) * 4 / 6,0)).
    
    ASSIGN insert-at = 1.
    DO b = 1 TO LENGTH(in-string) BY 6:
        ASSIGN
            PUT-BITS(sub,31,2) = INDEX(base64,SUBSTRING(in-string,b,1)) - 1
            PUT-BITS(sub,25,6) = INDEX(base64,SUBSTRING(in-string,b + 1,1)) - 1
            PUT-BITS(sub,19,6) = INDEX(base64,SUBSTRING(in-string,b + 2,1)) - 1
            PUT-BITS(sub,13,6) = INDEX(base64,SUBSTRING(in-string,b + 3,1)) - 1
            PUT-BITS(sub,7,6) = INDEX(base64,SUBSTRING(in-string,b + 4,1)) - 1
            PUT-BITS(sub,1,6) = INDEX(base64,SUBSTRING(in-string,b + 5,1)) - 1
            PUT-LONG(out-raw,insert-at) = sub
            insert-at = insert-at + 4
            .
    END.

    RETURN out-raw.
END FUNCTION.

